var __create = Object.create;
var __defProp = Object.defineProperty;
var __getProtoOf = Object.getPrototypeOf;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __markAsModule = (target) => __defProp(target, "__esModule", {value: true});
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, {get: all[name], enumerable: true});
};
var __reExport = (target, module2, desc) => {
  if (module2 && typeof module2 === "object" || typeof module2 === "function") {
    for (let key of __getOwnPropNames(module2))
      if (!__hasOwnProp.call(target, key) && key !== "default")
        __defProp(target, key, {get: () => module2[key], enumerable: !(desc = __getOwnPropDesc(module2, key)) || desc.enumerable});
  }
  return target;
};
var __toModule = (module2) => {
  return __reExport(__markAsModule(__defProp(module2 != null ? __create(__getProtoOf(module2)) : {}, "default", module2 && module2.__esModule && "default" in module2 ? {get: () => module2.default, enumerable: true} : {value: module2, enumerable: true})), module2);
};
__markAsModule(exports);
__export(exports, {
  conclude: () => conclude,
  finished: () => finished,
  getResult: () => getResult,
  inProgress: () => inProgress,
  isEffect: () => isEffect,
  isFlow: () => isFlow,
  isIterator: () => isIterator,
  isPromise: () => isPromise,
  whenFinished: () => whenFinished
});
var import_effects = __toModule(require("./effects"));
function isIterator(obj) {
  return !!obj && (typeof obj === "object" || typeof obj === "function") && typeof obj.next === "function" && typeof obj.throw === "function";
}
function isPromise(obj) {
  return !!obj && (typeof obj === "object" || typeof obj === "function") && typeof obj.then === "function" && typeof obj.catch === "function";
}
const isEffect = (effect) => Boolean(effect && effect[import_effects.TYPE]);
const isFlow = (it) => [isPromise, isEffect, isIterator].find((is) => is(it));
const runners = new Map([
  [isPromise, runPromise],
  [isEffect, runEffect],
  [isIterator, runIterator]
]);
const noop = () => {
};
const running = new WeakMap();
const resultCache = new WeakMap();
const finishWatchers = new WeakMap();
function inProgress(it) {
  return running.has(it);
}
function finished(it) {
  return resultCache.has(it);
}
function getResult(it) {
  return resultCache.get(it);
}
function whenFinished(it, callback) {
  if (!isFlow(it)) {
    callback({result: it});
    return noop;
  }
  if (resultCache.has(it)) {
    callback(resultCache.get(it));
    return noop;
  }
  let watchers = finishWatchers.get(it);
  if (!watchers)
    finishWatchers.set(it, watchers = new Set([callback]));
  else
    watchers.add(callback);
  return () => watchers.delete(callback);
}
function finalize(it, payload) {
  resultCache.set(it, payload);
  running.delete(it);
  for (let cb of finishWatchers.get(it) || [])
    cb(payload);
  finishWatchers.delete(it);
}
function conclude(it, callback) {
  const flowType = isFlow(it);
  if (!flowType) {
    callback(null, it);
    return noop;
  }
  if (resultCache.has(it)) {
    const {result, error, cancelled} = resultCache.get(it);
    if (cancelled)
      return noop;
    if (error)
      callback(error);
    else
      callback(null, result);
    return noop;
  }
  if (running.has(it)) {
    const subscribe2 = running.get(it);
    return subscribe2(callback);
  }
  const subscribers = new Set();
  const onConclude = (error, result) => {
    finalize(it, {error, result});
    for (let cb of subscribers)
      cb(error, result);
  };
  function subscribe(cb) {
    subscribers.add(cb);
    return function unsubscribe2() {
      subscribers.delete(cb);
      if (subscribers.size === 0) {
        finalize(it, {cancelled: true});
        cancel();
      }
    };
  }
  running.set(it, subscribe);
  const unsubscribe = subscribe(callback);
  const cancel = runners.get(flowType)(it, onConclude);
  return unsubscribe;
}
function runPromise(promise, callback) {
  let cancelled = false;
  promise.then((result) => !cancelled && callback(null, result)).catch((error) => !cancelled && callback(error));
  return () => cancelled = true;
}
function runEffect({[import_effects.TYPE]: type, context, fn, args}, callback) {
  try {
    switch (type) {
      case "CPS":
        return fn.call(context, ...args, callback);
      case "CPS_NO_CANCEL":
        let cancelled = false;
        fn.call(context, ...args, (error, result2) => !cancelled && callback(error, result2));
        return () => cancelled = true;
      case "CALL":
        const result = fn.apply(context, args);
        return conclude(result, callback);
      default:
        throw new Error("Unknown effect type " + type);
    }
  } catch (error) {
    callback(error);
    return noop;
  }
}
function runIterator(it, callback) {
  let cancel, step = 0;
  const setCancel = (j, fn) => {
    if (j >= step)
      cancel = fn;
  };
  function iterate(error, result) {
    try {
      let cancelled = false;
      setCancel(++step, () => cancelled = true);
      const {value, done} = error ? it.throw(error) : it.next(result);
      if (cancelled)
        return;
      setCancel(step, conclude(value, done ? callback : iterate));
    } catch (err) {
      callback(err);
    }
  }
  iterate();
  return () => cancel();
}
